<?php

declare(strict_types=1);

namespace Drupal\commerce_gnu_taler;

use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_price\Price;

/**
 * Taler backend order adapter.
 *
 * @phpstan-import-type CreateRedirectOrderParams from \Drupal\commerce_gnu_taler\TalerClient
 * @phpstan-import-type PostOrderRefundParams from \Drupal\commerce_gnu_taler\TalerClient
 */
class OrderAdapter {

  /**
   * Adapts a given payment entity to a GNU Taler order record.
   *
   * @param \Drupal\commerce_payment\Entity\PaymentInterface $payment
   *   The commerce payment entity.
   * @param string $returnUrl
   *   Location where a user agent is redirected when the payment was accepted.
   * @param string $cancelUrl
   *   Location where a user agent is redirected when the payment was cancelled.
   *
   * @return CreateRedirectOrderParams
   *   Returns params ready to pass to the create order call.
   */
  public function adaptCreateRedirectParams(PaymentInterface $payment, string $returnUrl, string $cancelUrl): array {
    $commerceOrder = $payment->getOrder();
    assert($commerceOrder !== NULL);

    $amount = $payment->getAmount();
    assert($amount !== NULL);

    return [
      'order' => [
        'amount' => $this->toRemoteCurrency($amount->getCurrencyCode()) . ':' . $amount->getNumber(),
        'summary' => (string) $commerceOrder->label(),
        'fulfillment_url' => $returnUrl,
        'order_id' => $commerceOrder->getOrderNumber(),
      ],
    ];
  }

  /**
   * Adapts a given payment entity to a GNU Taler refund request.
   *
   * @return PostOrderRefundParams
   *   Returns params ready to pass to the payment refund call.
   */
  public function adaptRefundRequest(PaymentInterface $payment, Price $amount): array {
    return [
      'refund' => $this->toRemoteCurrency($amount->getCurrencyCode()) . ':' . $amount->getNumber(),
      'reason' => '',
    ];
  }

  /**
   * Maps an ISO 4217 currency code to Taler backend currency key.
   */
  public function toRemoteCurrency(string $currencyCode): string {
    return match($currencyCode) {
      'KUD' => 'KUDOS',
      default => $currencyCode,
    };
  }

  /**
   * Maps Taler backend currency key to an ISO 4217 currency code.
   */
  public function toSystemCurrency(string $remoteCurrency): string {
    return match($remoteCurrency) {
      'KUDOS' => 'KUD',
      default => $remoteCurrency,
    };
  }

}
