<?php

declare(strict_types=1);

namespace Drupal\commerce_gnu_taler;

use GuzzleHttp\ClientInterface;

/**
 * Taler backend client.
 *
 * @phpstan-type CreateRedirectOrderParams array{
 *   'order': array{
 *     'amount': string,
 *     'summary': string,
 *     'fulfillment_url': string,
 *     'order_id': string
 *   }
 * }
 *
 * @phpstan-type CreateRedirectOrderResult array{
 *   'order_id': string,
 *   'token': string,
 * }
 *
 * @phpstan-type GetOrderStatusResult array{
 *   order_status: 'paid'|'claimed'|'unpaid',
 *   order_status_url: string,
 *   contract_terms?: array{
 *     amount: string,
 *     summary: string,
 *     order_id: string,
 *   }
 * }
 *
 * @phpstan-type PostOrderRefundParams array{'refund': string, 'reason': string}
 */
class TalerClient {

  /**
   * The Taler merchant protocol supported by this implementation.
   */
  public const MERCHANT_PROTOCOL_CURRENT = 18;

  /**
   * Number of protocol versions this library is backwards compatible with.
   */
  public const MERCHANT_PROTOCOL_AGE = 2;

  /**
   * Constructs the Taler backend client.
   */
  public function __construct(
    protected ClientInterface $httpClient,
    protected string $url,
    protected string $apiKey,
  ) {}

  /**
   * Returns true if the backend is protocol compatible.
   */
  public function verifyBackend(): bool {
    $response = $this->httpClient->request('GET', $this->url . 'config');
    if ($response->getStatusCode() !== 200) {
      return FALSE;
    }

    try {
      $configBody = json_decode((string) $response->getBody(), flags: JSON_OBJECT_AS_ARRAY | JSON_THROW_ON_ERROR);
    }
    catch (\JsonException $e) {
      return FALSE;
    }

    assert(is_array($configBody));
    if (!is_string($configBody['version'])) {
      return FALSE;
    }

    [$current, $revision, $age] = explode(':', $configBody['version'], 3);
    if (
      !is_numeric($current) ||
      !is_numeric($revision) ||
      !is_numeric($age) ||
      $age > $current
    ) {
      return FALSE;
    }

    if ($current > static::MERCHANT_PROTOCOL_CURRENT) {
      // Our implementation is too old!
      return FALSE;
    }

    if ($current < static::MERCHANT_PROTOCOL_CURRENT - static::MERCHANT_PROTOCOL_AGE) {
      // Merchant implementation is too old!
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Returns true if the backend supports the given currency.
   */
  public function verifyCurrency(string $expectedCurrency): bool {
    $response = $this->httpClient->request('GET', $this->url . 'config');
    if ($response->getStatusCode() !== 200) {
      return FALSE;
    }

    try {
      $configBody = json_decode((string) $response->getBody(), flags: JSON_OBJECT_AS_ARRAY | JSON_THROW_ON_ERROR);
    }
    catch (\JsonException $e) {
      return FALSE;
    }

    assert(is_array($configBody));
    if (!is_string($configBody['currency'])) {
      return FALSE;
    }

    return strcasecmp($expectedCurrency, $configBody['currency']) === 0;
  }

  /**
   * Creates a redirect order and returns the result.
   *
   * @param CreateRedirectOrderParams $params
   *   Redirect order params.
   *
   * @return CreateRedirectOrderResult|false
   *   Redirect result including order ID and token or false if call failed.
   *
   * @see https://docs.taler.net/core/api-merchant.html#post-[-instances-$INSTANCE]-private-orders
   */
  public function createRedirectOrder(array $params): array|FALSE {
    $response = $this->httpClient->request('POST', $this->url . 'private/orders', [
      'headers' => ['Authorization' => 'Bearer ' . $this->apiKey],
      'json' => $params,
    ]);
    if ($response->getStatusCode() !== 200) {
      return FALSE;
    }

    try {
      $orderBody = json_decode((string) $response->getBody(), flags: JSON_OBJECT_AS_ARRAY | JSON_THROW_ON_ERROR);
    }
    catch (\JsonException $e) {
      return FALSE;
    }

    assert(is_array($orderBody));
    assert(is_string($orderBody['order_id']));
    assert(is_string($orderBody['token']));
    return $orderBody;
  }

  /**
   * Returns the status for the given order id.
   *
   * @param string $remoteOrderId
   *   The remote order id.
   *
   * @return GetOrderStatusResult|false
   *   Order status result or false if call to backend failed.
   *
   * @see https://docs.taler.net/core/api-merchant.html#get-[-instances-$INSTANCE]-private-orders-$ORDER_ID
   */
  public function getOrderStatus(string $remoteOrderId): array|FALSE {
    $response = $this->httpClient->request('GET', $this->url . 'private/orders/' . $remoteOrderId, [
      'headers' => ['Authorization' => 'Bearer ' . $this->apiKey],
    ]);
    if ($response->getStatusCode() !== 200) {
      return FALSE;
    }

    try {
      $merchantOrderStatusResponse = json_decode(
        (string) $response->getBody(),
        flags: JSON_OBJECT_AS_ARRAY | JSON_THROW_ON_ERROR
      );
    }
    catch (\JsonException $e) {
      return FALSE;
    }

    assert(is_array($merchantOrderStatusResponse));
    assert(is_string($merchantOrderStatusResponse['order_status']));
    assert(in_array($merchantOrderStatusResponse['order_status'], ['paid', 'claimed', 'unpaid'], strict: TRUE));
    assert(is_string($merchantOrderStatusResponse['order_status_url']));
    return $merchantOrderStatusResponse;
  }

  /**
   * Starts a refund operation and returns an URI for the customers wallet.
   *
   * @param string $remoteOrderId
   *   GNU Taler order id.
   * @param array{'refund': string, 'reason': string} $refundRequest
   *   Refund request body.
   *
   * @return array{'taler_refund_uri': string, 'h_contract': string}|false
   *   The order refund result or false if the call to the backend failed.
   *
   * @see https://docs.taler.net/core/api-merchant.html#post-[-instances-$INSTANCE]-private-orders-$ORDER_ID-refund
   */
  public function postOrderRefund(string $remoteOrderId, array $refundRequest): array|FALSE {
    $response = $this->httpClient->request('POST', $this->url . 'private/orders/' . $remoteOrderId . '/refund', [
      'headers' => ['Authorization' => 'Bearer ' . $this->apiKey],
      'json' => $refundRequest,
    ]);
    if ($response->getStatusCode() !== 200) {
      return FALSE;
    }

    try {
      $merchantRefundResponse = json_decode(
        (string) $response->getBody(),
        flags: JSON_OBJECT_AS_ARRAY | JSON_THROW_ON_ERROR
      );
    }
    catch (\JsonException $e) {
      return FALSE;
    }

    assert(is_array($merchantRefundResponse));
    assert(is_string($merchantRefundResponse['taler_refund_uri']));
    assert(is_string($merchantRefundResponse['h_contract']));
    return $merchantRefundResponse;
  }

}
